unit QuatrisFrm;

interface

uses
  QuatrisCtrl, ActnList, Menus, Messages, ExtCtrls, Controls, Classes,
  DXClass, DXDraws, DXSounds;

type
  TRepaintPart = (rpScore, rpBackGround, rpBlocks, rpGameOver, rpHighScores,
    rpAgainScore, rpAgainBackGround, rpAgainBlocks, rpAgainGameOver, rpAgainHighScores);
  TRepaintParts = set of TRepaintPart;

  TView = (vNormal, vHighScores);

const
  rpAll = [rpScore .. rpHighScores];

const
  MaxHighScoreCount = 10;

type
  { TQuatrisHighScore }

  TQuatrisHighScore = class (TObject)
  private
    FScores: array[1..MaxHighScoreCount] of Integer;
    FNames: array[1..MaxHighScoreCount] of string;
    FHighScoreCount: Integer;
    FMustSave: Boolean;
    function GetScore(Index: Integer): Integer;
    function GetName(Index: Integer): string;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function IsHighScore(AScore: Integer): Boolean;
    procedure SetHighScore(AScore: Integer; const AName: string);

    property Score[Index: Integer]: Integer read GetScore;
    property Name[Index: Integer]: string read GetName;
    property HighScoreCount: Integer read FHighScoreCount;
  end;

  { TQuatrisForm }

  TQuatrisForm = class(TDXForm)
    DXDraw: TDXDraw;
    DXTimer: TTimer;
    ImageList: TDXImageList;
    ActionList: TActionList;
    NewGame: TAction;
    FullScreen: TAction;
    HighScores: TAction;
    DXWaveList: TDXWaveList;
    DXSound: TDXSound;
    SoundEnabled: TAction;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure DXTimerTimer(Sender: TObject);
    procedure DXDrawInitializeSurface(Sender: TObject);
    procedure NewGameExecute(Sender: TObject);
    procedure FullScreenExecute(Sender: TObject);
    procedure DXDrawInitializing(Sender: TObject);

    procedure RepaintAll(Sender: TObject);
    procedure DXDrawMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure HighScoresExecute(Sender: TObject);
    procedure SoundEnabledExecute(Sender: TObject);
  private
    { Private declarations }
    PepijnIndex: Integer;
    FDoubleBuffer: Boolean;
    FGameOverIndex: Integer;
    FHighScore: TQuatrisHighScore;
    FView: TView;
    FRepaintParts: TRepaintParts;
    procedure ScoreChange(Sender: TObject);
    procedure FormRepaint;
    procedure OnNewGame(Sender: TObject);
    procedure OnGameOver(Sender: TObject);
    procedure SetView(AView: TView);
    procedure OnFallBlock(Sender: TObject);
    procedure OnSetCurrentBlock(Sender: TObject);
    procedure ApplIdle(Sender: TObject; var Done: Boolean);
  public
    { Public declarations }
    procedure PlaySound(Index: Integer);

    property HighScore: TQuatrisHighScore read FHighScore;
    property View: TView read FView write SetView;
    property RepaintParts: TRepaintParts read FRepaintParts write FRepaintParts;
  end;

// Form and score repaint are request when something has changed and handled
// when the program goes idle.

var
  QuatrisForm: TQuatrisForm;
  AQuatris: TQuatris;
  BlokSize: Integer;
  BlokTop: Integer;
  BlokLeft: Integer;

const
  ScoreHeight = 17;
  ScoreWidth = 12;
  ScoreBlockWidth = 4 * ScoreWidth;
  ScoreImgIndex = 8;
  GameOverImgIndex = 18;
  NewGameImgIndex = 19;
  FullScreenImgIndex = 20;
  DoNotUseImgIndex = 21;
  FullScreenDownImgIndex = 22;
  HighScoresImgIndex = 23;
  HighScoresDownImgIndex = 24;
  SoundImgIndex = 25;
  SoundDownImgIndex = 26;

  FallSoundIndex = 0;
  SetCurrentBlockSoundIndex = 1;
  StartupSoundIndex = 2;
  ScoreChangeSoundIndex = 3;
  ButtonClickSoundIndex = 4;

implementation

uses
  Registry, Dialogs, Windows, Graphics, Forms, SysUtils;

{$R *.DFM}

const
  RegKey: string = 'Software\Python\Quatris';
  RegOptions: string = 'Options';
  RegFullScreen: string = 'FullScreen';
  RegSound: string = 'Sound';

procedure TQuatrisForm.FormCreate(Sender: TObject);
begin
  // Read the options info.
  with TRegIniFile.Create(RegKey) do
  try
    FullScreen.Checked := ReadBool(RegOptions, RegFullScreen, False);
    SoundEnabled.Checked := ReadBool(RegOptions, RegSound, True);
  finally
    Free;
  end;

  // If fullscreen is set also make the application fullscreen.
  if FullScreen.Checked then
  begin
    DXDraw.Finalize;
    DXDraw.Options := DXDraw.Options + [doFullScreen];
    DXDraw.Initialize;
  end;

  Randomize;

  // Create the Quatris object.
  AQuatris := TQuatris.Create;
  AQuatris.OnSetCurrentBlok := OnSetCurrentBlock;
  AQuatris.OnFallBlock := OnFallBlock;
  AQuatris.OnScoreChange := ScoreChange;
  AQuatris.OnGameOver := OnGameOver;
  AQuatris.OnNewGame := OnNewGame;

  // Make ImageList ready.
  ImageList.Items.MakeColorTable;
  DXDraw.ColorTable := ImageList.Items.ColorTable;
  DXDraw.DefColorTable := ImageList.Items.ColorTable;

  Application.OnIdle := ApplIdle;
  FHighScore := TQuatrisHighScore.Create;
  RepaintParts := rpAll;

  PlaySound(StartupSoundIndex);
end;

procedure TQuatrisForm.FormDestroy(Sender: TObject);
begin
  // Save the options info.
  with TRegIniFile.Create(RegKey) do
  try
    WriteBool(RegOptions, RegFullScreen, FullScreen.Checked);
    WriteBool(RegOptions, RegSound, SoundEnabled.Checked);
  finally
    Free;
  end;

  AQuatris.Free;
  FHighScore.Free;
end;

procedure TQuatrisForm.DXTimerTimer(Sender: TObject);
begin
  if AQuatris.AGameOver then
  begin
    if FGameOverIndex < 18 then
    begin
      Inc(FGameOverIndex);
      RepaintParts := RepaintParts + [rpGameOver, rpBlocks];
    end;
  end
  else
    AQuatris.FallBlok;
end;

const
  Pepijn = 'PEPIJN';

procedure TQuatrisForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  // check for typing pepijn
  if (not AQuatris.AGameOver) and (Key = Ord(Pepijn[PepijnIndex + 1])) then
  begin
    Inc(PepijnIndex);
    if PepijnIndex = Length(Pepijn) then
    begin
      AQuatris.Score := AQuatris.Score + 10;
      PepijnIndex := 0;
    end;
  end
  else
    PepijnIndex := 0;

  if Shift <> [] then Exit;
  case Key of
    37: AQuatris.BlokToLeft;
    32, 38: AQuatris.Rotate;
    39: AQuatris.BlokToRight;
    40: if AQuatris.CanFall then
        begin
          DXTimer.Enabled := False;
          AQuatris.FallBlok;
          DXTimer.Enabled := True;
        end;
  end;
end;

procedure TQuatrisForm.DXDrawInitializeSurface(Sender: TObject);
begin
  // Calculate the bigest possible size, with a border of 5.
  BlokSize := (DXDraw.SurfaceHeight -10) div 20;
  if BlokSize > (DXDraw.SurfaceWidth -10) div 22 then
    BlokSize := (DXDraw.SurfaceWidth -10) div 22;
  // Calculate the distance to the top and left side.
  BlokTop := (DXDraw.SurfaceHeight - BlokSize * 20) div 2;
  BlokLeft := (DXDraw.SurfaceWidth - BlokSize * 10) div 2;
  // Repaint.
  RepaintParts := rpAll;
end;

procedure TQuatrisForm.NewGameExecute(Sender: TObject);
begin
  PlaySound(ButtonClickSoundIndex);
  AQuatris.NewGame;
end;

procedure TQuatrisForm.FullScreenExecute(Sender: TObject);
begin
  PlaySound(ButtonClickSoundIndex);
  
  FullScreen.Checked := not FullScreen.Checked;
  DXDraw.Finalize;
  if FullScreen.Checked then
  begin
    DXDraw.Options := DXDraw.Options + [doFullScreen];
  end
  else
  begin
    DXDraw.Options := DXDraw.Options - [doFullScreen];
  end;
  DXDraw.Initialize;
  RepaintParts := rpAll;
end;

procedure TQuatrisForm.DXDrawInitializing(Sender: TObject);
begin
  if doFullScreen in DXDraw.Options then
  begin
    BorderStyle := bsNone;
    FDoubleBuffer := True;
  end
  else
  begin
    FDoubleBuffer := False;
    BorderStyle := bsSizeable;
  end;
end;

procedure TQuatrisForm.ScoreChange(Sender: TObject);
begin
  // Adjust the interval, using the score.
  DXTimer.Interval := 50000 div (100 + AQuatris.Score);
  // Request a score repaint.
  RepaintParts := RepaintParts + [rpScore];

  PlaySound(ScoreChangeSoundIndex);
end;

procedure TQuatrisForm.FormRepaint;

  procedure RepaintAlreadyFallenBlocks;
  var
    AKolom, ARij, I: Integer;
  begin
    // Draw the already fallen blocks.
    // Draw the  background picture.
    ImageList.Items[0].StretchDraw(DXDraw.Surface,
        Rect(BlokLeft, BlokTop, BlokLeft + 10 * BlokSize, BlokTop + 20 * BlokSize), 0);
    for AKolom := 0 to 9 do
      for ARij := 0 to 19 do
      begin
        I := AQuatris.Bord[ARij, AKolom];
        if (I <> 0) and (I <> 2) then
          ImageList.Items[I].DrawRotate(DXDraw.Surface,
            BlokLeft + AKolom * BlokSize, BlokTop + ARij * BlokSize, BlokSize,
               BlokSize, 0, 0, 0, 0);
      end;
  end;

  procedure RepaintFallingBlock;
  var
    ARij, AKolom: Integer;
  begin
    // Draw the falling block.
    ARij := AQuatris.BlokKleur;
    for AKolom := 0 to 3 do
      with AQuatris.CurrentBlok[AKolom] do
        ImageList.Items[ARij].DrawRotate(DXDraw.Surface, BlokLeft + Kolom*BlokSize,
           BlokTop + Rij*BlokSize, BlokSize, BlokSize, 0, 0, 0, 0);
  end;

  procedure RepaintScore;
  var
    Score, Left: Integer;
  begin
    // Overwrite the old score
    DXDraw.Surface.FillRect(Rect(BlokLeft + 12 * BlokSize, BlokTop + 10 * BlokSize,
        BlokLeft + 16 * BlokSize, BlokTop + 10 * BlokSize + 17 * BlokSize div 12), 0);
    // write the new score.
    Left := BlokLeft + 15 * BlokSize;
    Score := AQuatris.Score;
    repeat
      ImageList.Items[ScoreImgIndex + Score mod 10].StretchDraw(DXDraw.Surface,
        Rect(Left, BlokTop + 10 * BlokSize,
             Left + BlokSize, BlokTop + 10 * BlokSize + 17 * BlokSize div 12), 0);
      Left := Left - BlokSize;
      Score := Score div 10;
    until Score = 0;
  end;

  procedure RepaintGameOver;
  var
    I: Integer;
  begin
    // Paint the game over blok.
    I := FGameOverIndex;
    ImageList.Items[GameOverImgIndex].StretchDraw(DXDraw.Surface,
       Rect(BlokLeft, BlokTop + I * BlokSize,
            BlokLeft + 10 * BlokSize, BlokTop + (I +2) * BlokSize), 0);
  end;

  procedure RepaintBackGround;
  begin
    // Make everything black.
    DXDraw.Surface.Fill(0);
    // Draw the NewGame button.
    ImageList.Items[NewGameImgIndex].StretchDraw(DXDraw.Surface,
       Rect(BlokSize, BlokSize,
            6 * BlokSize, BlokSize * 2), 0);
    // Draw the fullscreen button.
    if FullScreen.Checked then
      ImageList.Items[FullScreenDownImgIndex].StretchDraw(DXDraw.Surface,
         Rect(BlokSize, BlokSize * 3,
              6 * BlokSize, BlokSize * 4), 0)
    else
      ImageList.Items[FullScreenImgIndex].StretchDraw(DXDraw.Surface,
         Rect(BlokSize, BlokSize * 3,
              6 * BlokSize, BlokSize * 4), 0);
    // Draw the HighScore button
    if View = vHighScores then
      ImageList.Items[HighScoresDownImgIndex].StretchDraw(DXDraw.Surface,
        Rect(BlokSize, BlokSize * 5, 6 * BlokSize, BlokSize * 6), 0)
    else
      ImageList.Items[HighScoresImgIndex].StretchDraw(DXDraw.Surface,
        Rect(BlokSize, BlokSize * 5, 6 * BlokSize, BlokSize * 6), 0);
    // Draw the DoNotUse button.
    ImageList.Items[DoNotUseImgIndex].StretchDraw(DXDraw.Surface,
       Rect(BlokSize, BlokSize * 9,
            6 * BlokSize, BlokSize * 10), 0);
    // Draw the SoundEnabled button.
    if SoundEnabled.Checked then
      ImageList.Items[SoundDownImgIndex].StretchDraw(DXDraw.Surface,
        Rect(BlokSize, BlokSize * 7, 6 * BlokSize, BlokSize * 8), 0)
    else
      ImageList.Items[SoundImgIndex].StretchDraw(DXDraw.Surface,
        Rect(BlokSize, BlokSize * 7, 6 * BlokSize, BlokSize * 8), 0);
  end;

  procedure RepaintHighScores;
  var
    Canvas: TDirectDrawSurfaceCanvas;
    ScoreText: string;
    ScoreWidth: Integer;
    I: Integer;
  begin
    // Paint the HighScores.
    Canvas := DXDraw.Surface.Canvas;
    try
      Canvas.Brush.Color :=  clBlack;
      Canvas.Font.Color := clRed;
      Canvas.Font.Height := BlokSize;
      for I := 1 to HighScore.HighScoreCount do
      begin
        Canvas.TextRect(Rect(BlokLeft, BlokTop + BlokSize * (I - 1), BlokLeft + BlokSize * 10, BlokTop + BlokSize * I),
          BlokLeft, BlokTop + BlokSize * (I - 1), Format('%d: %s', [I, HighScore.Name[I]]));
        ScoreText := IntToStr(HighScore.Score[I]);
        ScoreWidth := Canvas.TextWidth(ScoreText);
        Canvas.TextRect(Rect(BlokLeft + BlokSize * 10 - ScoreWidth, BlokTop + BlokSize * (I - 1), BlokLeft + BlokSize * 10, BlokTop + BlokSize * I),
          BlokLeft + BlokSize * 10 - ScoreWidth, BlokTop + BlokSize * (I - 1), ScoreText);
      end;
    finally
      Canvas.Release;
    end;
  end;

begin
  if not DXDraw.CanDraw then
    Exit;

  if (rpBackGround in RepaintParts) or (rpAgainBackGround in RepaintParts) then
  begin
    RepaintBackGround;
    // blcoks and score also need repainted when the background is repainted.
    RepaintParts := RepaintParts + [rpBlocks, rpScore];
  end;

  case View of
    vNormal:
      begin
      if (rpBlocks in RepaintParts) or (rpAgainBlocks in RepaintParts) then
      begin
        RepaintAlreadyFallenBlocks;
        RepaintFallingBlock;
      end;

      // Print the new score on the score surface.
      if (rpScore in RepaintParts) or (rpAgainScore in RepaintParts) then
        RepaintScore;

      if AQuatris.AGameOver and ((rpGameOver in RepaintParts) or (rpAgainGameOver in RepaintParts)) then
        RepaintGameOver;
      end;
    vHighScores:
      begin
      if (rpBackGround in RepaintParts) or (rpAgainBackGround in RepaintParts) then
        RepaintHighScores;
      end;
  end;

  // Flip the image.
  DXDraw.Flip;

  if FDoubleBuffer then
    RepaintParts := TRepaintParts(Word((Word(RepaintParts) mod 64) shl 4))
  else
    RepaintParts := [];
end;

procedure TQuatrisForm.ApplIdle(Sender: TObject; var Done: Boolean);
begin
  if RepaintParts <> [] then
  begin
    FormRepaint;
    Done := False;
  end;
end;

procedure TQuatrisForm.RepaintAll(Sender: TObject);
begin
  RepaintParts := rpAll;
end;

procedure TQuatrisForm.DXDrawMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (X >= BlokSize) and (X <= 6 * BlokSize) then
  begin
    if (Y >= BlokSize) and (Y <= 2 * BlokSize) then
      NewGame.Execute;
    if (Y >= 3 * BlokSize) and (Y <= 4 * BlokSize) then
      FullScreen.Execute;
    if (Y >= 5 * BlokSize) and (Y <= 6 * BlokSize) then
      HighScores.Execute;
    if (Y >= 7 * BlokSize) and (Y <= 8 * BlokSize) then
      SoundEnabled.Execute;
    if (Y >= 9 * BlokSize) and (Y <= 10 * BlokSize) then
      Close;
  end;
end;

procedure TQuatrisForm.OnNewGame(Sender: TObject);
begin
  // Update things.
  FGameOverIndex := 0;
  View := vNormal;
end;

procedure TQuatrisForm.OnGameOver(Sender: TObject);
resourcestring
  SHighScoreCaption = 'You have a high score !?';
  SHighScorePrompt = 'Enter your name:';
begin
  // Check for highscore.
  if HighScore.IsHighScore(AQuatris.Score) then
    HighScore.SetHighScore(AQuatris.Score,
       InputBox(SHighScoreCaption, SHighScorePrompt, ''));
  // Let the game over block fall.
  DXTimerTimer(nil);
end;

procedure TQuatrisForm.SetView(AView: TView);
begin
  if FView <> AView then
  begin
    FView := AView;
    // Repaint everything.
    RepaintParts := rpAll;
  end;
end;

procedure TQuatrisForm.HighScoresExecute(Sender: TObject);
begin
  PlaySound(ButtonClickSoundIndex);

  if View = vHighScores then
    View := vNormal
  else
    View := vHighScores;
end;

procedure TQuatrisForm.SoundEnabledExecute(Sender: TObject);
begin
  (Sender as TAction).Checked := not (Sender as TAction).Checked;
  
  PlaySound(ButtonClickSoundIndex);
end;

procedure TQuatrisForm.OnFallBlock(Sender: TObject);
begin
  PlaySound(FallSoundIndex);
  // Repaint blocks.
  RepaintParts := RepaintParts + [rpBlocks];
end;

procedure TQuatrisForm.OnSetCurrentBlock(Sender: TObject);
begin
  PlaySound(SetCurrentBlockSoundIndex);
  // Repaint block.
  RepaintParts := RepaintParts + [rpBlocks];
end;

procedure TQuatrisForm.PlaySound(Index: Integer);
begin
  if SoundEnabled.Checked then
    DXWaveList.Items[Index].Play(False);

  RepaintParts := RepaintParts + [rpBackground];
end;

{ TQuatrisHighScore }

const
  RegSection: string = 'HighScore';
  RegCount: string = 'Count';
  RegName: string = 'Name%d';
  RegScore: string = 'Score%d';
  {$IFNDEF NoOldHighScore}
  OldRegName: string = 'Name';
  OldRegScore: string = 'Score';
  {$ENDIF}

constructor TQuatrisHighScore.Create;
var
  RegIni: TRegIniFile;
  I: Integer;
begin
  inherited Create;
  RegIni := TRegIniFile.Create(RegKey);
  try
    FHighScoreCount := RegIni.ReadInteger(RegSection, RegCount, 0);
    for I := 1 to HighScoreCount -1 do
    begin
      FNames[I] := RegIni.ReadString(RegSection, Format(RegName, [I]), '');
      FScores[I] := RegIni.ReadInteger(RegSection, Format(RegScore, [I]), 0);
    end;

    {$IFNDEF NoOldHighScore}
    // Support for old HighScore values.
    if HighScoreCount = 0 then
    begin
      FNames[1] := RegIni.ReadString(RegSection, OldRegName, '');
      FScores[1] := RegIni.ReadInteger(RegSection, OldRegScore, 0);
      RegIni.DeleteKey(RegSection, OldRegName);
      RegIni.DeleteKey(RegSection, OldRegScore);
      if FScores[1] <> 0 then
      begin
        FHighScoreCount := 1;
        FMustSave := True;
      end;
    end;
    {$ENDIF}
  finally
    RegIni.Free;
  end;
end;

destructor TQuatrisHighScore.Destroy;
var
  RegIni: TRegIniFile;
  I: Integer;
begin
  if FMustSave then
  begin
    RegIni := TRegIniFile.Create(RegKey);
    try
      RegIni.WriteInteger(RegSection, RegCount, HighScoreCount);
      for I := 1 to HighScoreCount do
      begin
        RegIni.WriteString(RegSection, Format(RegName, [I]),  Name[I]);
        RegIni.WriteInteger(RegSection, Format(RegScore, [I]), Score[I]);
      end;
    finally
      RegIni.Free;
    end;
  end;
  inherited Destroy;
end;

function TQuatrisHighScore.GetScore(Index: Integer): Integer;
begin
  Result := FScores[Index];
end;

function TQuatrisHighScore.GetName(Index: Integer): string;
begin
  Result := FNames[Index];
end;

function TQuatrisHighScore.IsHighScore(AScore: Integer): Boolean;
begin
  Result := AScore > Score[MaxHighScoreCount];
end;

procedure TQuatrisHighScore.SetHighScore(AScore: Integer; const AName: string);
var
  I, J: Integer;
resourcestring
  SNotAHighScoreError = 'Invalid HighScore call, HighScore isn''t a HighScore';
begin
  // Search the place where the item should be inserted.
  I := 1;
  while (I <= HighScoreCount) and (Score[I] >= AScore) do
    I := I + 1;
  if I > MaxHighScoreCount then
    raise Exception.Create(SNotAHighScoreError);

  // Insert the new highScore.
  if FHighScoreCount < MaxHighScoreCount then
    FHighScoreCount := FHighScoreCount + 1;
  for J := HighScoreCount downto I + 1 do
  begin
    FScores[J] := FScores[J - 1];
    FNames[J] := FNames[J - 1];
  end;
  FScores[I] := AScore;
  FNames[I] := AName;
  FMustSave := True;
end;

end.
 